#!/usr/bin/python3

import numpy as np
from scipy import stats
import sys


def getdata(fname):
    data = {}
    with open(fname, 'r') as f:
        for l in f:
            if "function" in l or "====" in l:
                continue
            op, args, result = l.split()
            if op in data:
                if args in data[op]:
                    data[op][args].append(float(result))
                else:
                    data[op][args] = [float(result)]
            else:
                data[op] = {args: [float(result)]}

    return data


def avgdata(data):
    for op in data:
        for args in data[op]:
            vals = data[op][args]
            data[op][args] = (np.average(vals), np.std(vals))
    return data


def getname(op):
    if 'div' in op:
        return "Division"
    elif 'mul' in op:
        return "Multiplication"
    elif 'sqrt' in op:
        return "Square root"
    elif 'add' in op:
        return "Addition"
    elif 'sub' in op:
        return "Subtraction"
    else:
        return "UNKNOWN"


def gettype(op):
    if '_dp' in op:
        return "double precision"
    elif '_sp' in op:
        return "single precision"
    else:
        return "UNKNOWN"


replacements = {'10000000000.0': '1e10', '9.99988867183e-321': '1e-320'}


def replace_value(value):
    if value in replacements:
        return replacements[value]
    else:
        return value


def print_op(op, values, data, name, cap):
    label = '''tbl_''' + str(op).replace("drag_", "") + '''_''' + name

    header = '''%<*''' + label.replace('_', '') + '''>
\\begin{figure*}[ht]
\\centering
\\begin{tabular}{c  ''' + "c " * len(values) + '''}
\\toprule'''
    if "div" in op:
        header = header + '''
\\multirow{3}{*}{Dividend} & \\multicolumn{''' + str(len(values)) + '''}{c}{Divisor}\\\\
'''
    mider = '''\cmidrule(r){2-''' + str(len(values) + 1) + '''}
& \\multicolumn{''' + str(len(values)) + '''}{c}{Cycle count}\\\\'''

    print(header)
    for v in values:
        print(" & " + replace_value(v), end=' ')
    print("\\\\", end=' ')

    print(mider)
    baseline = stats.mode([round(x)
                          for (x, y) in [data[op][args]
                                         for args in data[op]]],
                          axis=None)[0][0]
    current_a = ""

    for args in sorted(data[op].keys()):
        a, b = args.split(',')
        if a in values and b in values:
            if a != current_a:
                if current_a != "":
                    print('''\\\\
''', end=' ')
                current_a = a
                print(
                    '''\\multicolumn{1}{c|}{''' + replace_value(a) + "} ", end=' ')
            cell_print = "& "
            if(data[op][args][0] - baseline >= 1):
                cell_print = cell_print + " \\highcell{}"
            elif(data[op][args][0] - baseline <= -1):
                cell_print = cell_print + " \\lowcell{}"
            if(data[op][args][1] > 1.0):
                cell_print = cell_print + " \\bf{"
            cell_print = cell_print + "%.2f" % (data[op][args][0])
            if(data[op][args][1] > 1.0):
                cell_print = cell_print + "}"

            print(cell_print, end=' ')


    # Ending of the table
    print('''\\\\
\\bottomrule
\\end{tabular}
\\caption{''' + getname(op) + ''' timing for ''' + gettype(op) + ''' floats on ''' + cap + '''}
\\label{fig:''' + label + '''}
\\end{figure*}
%</''' + label.replace('_', '') + '''>
''')


def print_op_sqrt(op, values, data, name, cap):
    label = '''tbl_''' + str(op).replace("drag_", "") + '''_''' + name

    header = '''%<*''' + label.replace('_', '') + '''>
\\begin{figure}[ht]
\\centering
\\begin{tabular}{c | c}
'''
    mider = '''
Operand & Cycle count\\\\
\\midrule
'''

    print(header)

    print(mider)
    baseline = stats.mode([round(x)
                          for (x, y) in [data[op][args] for args in data[op]]], axis=None)[0][0]
    current_a = ""

    for args in sorted(data[op].keys()):
        a = args
        if a in values:
            cell_print = replace_value(a) + " & "
            if(data[op][args][0] - baseline >= 1):
                cell_print = cell_print + " \\highcell{}"
            elif(data[op][args][0] - baseline <= -1):
                cell_print = cell_print + " \\lowcell{}"
            if(data[op][args][1] > 1.0):
                cell_print = cell_print + " \\bf{"
            cell_print = cell_print + "%.2f" % (data[op][args][0])
            if(data[op][args][1] > 1.0):
                cell_print = cell_print + "}"

            print(cell_print + "\\\\")


    # Ending of the table
    print('''\\\\
\\end{tabular}
\\caption{''' + getname(op) + ''' timing for ''' + gettype(op) + ''' floats on ''' + cap + '''}
\\label{fig:''' + label + '''}
\\end{figure}
%</''' + label.replace('_', '') + '''>
''')


# ORDER MATTERS
values_dp = ['0.0', '1.0', '10000000000.0', '1e+200'  # , '1e+30', '1e-30'
             , '1e-300'  # , '1e-41'
             , '1e-42'  # , '255'
             , '256', '257', '9.99988867183e-321']
values_sp = ['0.0', '1.0', '10000000000.0', '1e+30', '1e-30', '1e-41', '1e-42'  # , '255'
             , '256', '257']

if(len(sys.argv) > 2):
    name = sys.argv[2]
else:
    name = ""

if(len(sys.argv) > 3):
    cap = sys.argv[3]
    print(cap)
else:
    cap = name

data = avgdata(getdata(sys.argv[1]))
for op in data:

    if("_dp" in op):
        values = values_dp
    else:
        values = values_sp

    if("sqrt" in op):
        print_op_sqrt(op, values, data, name, cap)
    else:
        print_op(op, values, data, name, cap)
